%
%Sound Card Setup (Octave/Matlab)
%
%Play stimulus out soundcard (both L and R channels),
%simulataneously record L and R intputs.
%View spectrum of inputs in "realtime".
%The program operates in a loop, constantly updating the spectrum and other
%displays.
%
%This program may be used to check sound card (loop L and R outputs
%to inputs).  This tests the sound card, eg., to make sure levels are not
%clipping, etc.
%
%It can also be used to test an audio device (amplifier, effects pedal, etc)
%or an entire audio chain (amp>speaker>air>microphone>preamp) by inserting the
%the "device under test" (DUT) in the right channel.
%
%The system is tested with a *periodic sequence* of length N.
%The system is stimulated with (two or more) periods of data.
%Towards the end of the stimulus exactly one period is extracted.
%This is done to stabilize the sytems to transients.  In our case
%just two periods are used.  The timing and debugging of this code
%are described below.
%
%N.B.: by using a periodic stimulus, we solve two problems.
%
%First, window effects are avoided. That is, a sinusoial stimulus
%(with an integer number of periods) will show up in a single bin
%of the DFT of length N.
%Periodic stimuli may be any type, inlcuding single sine wave,
%superposition of tones, white or pink noise, etc.
%
%The second reason to use periodic signals is to avoid timing issues.
%As long as once period is captured, it does not matter exactly where we
%start and end capture, as long as we are within the active data window.
%
%
%Notes on the User interface implementation are included in the comments below.
%
%Created Jul 1, 2016.
%Richard Mann
%Computer Science
%University of Waterloo
%
%TODO:
%-make this work with National Instruments myDAQ (Matlab/Windows only)
%
%-build proper GUI
close all; clear;
%graphics_toolkit fltk

if exist('octave_config_info'),
  % Octave specific code
  pkg load signal
  % Matlab sepcific code, add ELSE case
end %if octave

% Timing controls. Adjust for your operating system:
% NP: Number of Stimulus periods (set N=1.0 for testing).
% Pofs: Offset, start of sampling, in periods. (set Pofs=0 testing) 
% Tdelay: Time delay, in seconds, from record start to play start (0 for test)
if ispc
  NP=2.0; Pofs=0.75; Tdelay=0.31;
elseif isunix
  NP=2.5; Pofs=0.75; Tdelay=0.05;
elseif ismac
  NP=2.0; Pofs=0.5; Tdelay=0.0;
else
  error('Unrecognized Operating System.');
end

% Default appearance for plots
set(0,'DefaultLineLineWidth',1.5);
set(0,'DefaultAxesFontSize',12);
set(0,'DefaultAxesFontWeight','bold')
set(0,'DefaultAxesLineWidth',1.5)

% Comment/Uncomment below for Windows/Displays.
% The main loop below checks if figure Handles exist, and if so,
% updates those plots.  Set the default position/layout below.
figStimulus=figure; set(gcf,'units','normalized','doublebuffer','on'); p=get(gcf,'position');
drawnow;
set(gcf,'position',[0 1-p(4) p(3) p(4)]); % NorthWest

figInput=figure; set(gcf,'units','normalized','doublebuffer','on'); p=get(gcf,'position');
  set(gcf,'position',[0 0 p(3) p(4)]); % SouthWest

figSpectrum=figure; set(gcf,'units','normalized','doublebuffer','on'); p=get(gcf,'position');
  set(gcf,'position',[1-p(3) 1-p(4) p(3) p(4)]); % northeast 
figPhase=figure; set(gcf,'units','normalized','doublebuffer','on'); p=get(gcf,'position');
  set(gcf,'position',[1-p(3) 0.5-p(4)/2 p(3) p(4)]); % south
figTF=figure; set(gcf,'units','normalized','doublebuffer','on'); p=get(gcf,'position');
  set(gcf,'position',[1-p(3) 0 p(3) p(4)]); % southeast
%figImpulse=figure; set(gcf,'units','normalized','doublebuffer','on'); p=get(gcf,'position');
%  set(gcf,'position',[0.5-p(3)/2 1-p(4) p(3) p(4)]); % north
%figScatter=figure; set(gcf,'units','normalized','doublebuffer','on'); p=get(gcf,'position');
%  set(gcf,'position',[0.5-p(3)/2 0 p(3) p(4)]); % south

% This line makes sure all windows come up before program opens dialog
drawnow;

% The 'simplest possible' GUI.
% No callbacks are used.  Instead, "Radio Buttons" are used to set status
% variables ('value' field of the button interface,1=true, 0=fales).
% The main program loop below looks at the status of these buttons.
% All GUI elements are in a single window.
fh=figure; title('SoundCardSetup GUI');
set(gcf,'units','normalized');
set(gcf,'visible','on');
axis off;
% GUI elements are packed into window.  Just use Radio Buttons now.
rINPUT=uicontrol('style','radiobutton','string','STIMULUS' ,'units','normalized','position',[.1 .9 .2 .1],'value',0);
rRUN=uicontrol('style','radiobutton','string','RUN/STOP','units','normalized','position',[.1 .7 .2 .1],'value',0);
rWRITE=uicontrol('style','radiobutton','string','SAVE','units','normalized','position',[.1 .5 .2 .1],'value',0);
rEXIT=uicontrol('style','radiobutton','string','EXIT','units','normalized','position',[.1 .1 .2 .1],'value',0);

% Slightly more GUI.
% Level controls set a global variable 'UserData' in the window 0 (console).
% This is a string field read by the main loop below.
sLvol = uicontrol('Style','slider','units','normalized','Position',[0.5 0.1 0.05 0.8],'value',.25, 'min',0, 'max',1,'Callback',@(f,e)(set(0,'UserData','VOL')));
sRvol = uicontrol('Style','slider','units','normalized','Position',[0.6 0.1 0.05 0.8],'value',.25, 'min',0, 'max',1,'Callback',@(f,e)(set(0,'UserData','VOL')));
rRinvert = uicontrol('Style','radiobutton','string','INV','units','normalized','Position',[0.6 0 0.1 0.1],'value',0,'Callback',@(f,e)(set(0,'UserData','VOL')));
drawnow;

% Intial state,
set(rINPUT,'value',1); % start reading stimulus


% Program loop, will change behaviour based on above radio buttons.
while (1),

if get(rINPUT,'value')
% Set stimulus file, also calibration levels dB1 and dB2.
% Normally left and right channels will be equal, but we may also
% have quadrature or other signals in future...
fn=uigetfile('*.wav', 'Stimulus file (stereo WAV)');


[sig,fs]=audioread(fn); % "sig" is stimulus signal (stereo)
N=size(sig,1); ns=2*N;
Nbit=16; % PCM 16 bit
dB1=20*log10(2^(Nbit-1)); % ADC/DAC
dB2=20*log10(N/2); % Normalization by # of points in DFT

set(rINPUT,'value',0); % Input state ends
set(0,'UserData','VOL'); % Change volume event
end %if rINPUT

% The level/Volume controls are sliders, state is transfered by GUI field
% 'UserData'
if strcmp(get(0,'UserData'),'VOL')

% Generate stimulus
Ts=NP*N/fs; % Stimulus time
Ns=round(Ts*fs); % Stimulus samples
y=repmat(sig,ceil(Ns/N),1); % create longer record and truncate to Ns
y=y(1:Ns,:);

% Left and Right volume from sliders in GUI.
Lvol=get(sLvol,'value'); Rvol=get(sRvol,'value');
if get(rRinvert,'value') Rvol=-Rvol; end; % flip sign
disp(sprintf('Levels: %1.2f %1.2f', Lvol, Rvol));
y(:,1)=y(:,1)*Lvol;
y(:,2)=y(:,2)*Rvol;

% Create player and recorder objects once
rec = audiorecorder(fs,16,2);
ply = audioplayer(y,fs,16);


% Now move on to Running state...
set(0,'UserData',[]);
set(rRUN,'value',1);
end %if VOL_CHANGED

if get(rRUN,'value'),


play(ply); %this should play and allow immediate further execution
recordblocking(rec,Ts+Tdelay); % waits until all samples recorded

z = getaudiodata(rec,'double');% this retrieves the recorded data
ns=size(z,1); % total number of samples acquired

% select N samples, part way in sequence
Nofs=round(Pofs*N); % offset in samples.
Tofs=Nofs/fs; % offset in seceonds
Ndelay=round(Tdelay*fs);
t=((1:N)-1)'/fs;
ref=z(Ndelay+Nofs+(1:N),1);
dut=z(Ndelay+Nofs+(1:N),2);
t1=Tdelay+Tofs; t2=t1+N/fs;

if exist('figStimulus'),
set(0,'CurrentFigure',figStimulus);
clf; hold on; grid on;
xlim([0 Ns/fs]); ylim([-1.05 1.05]);
title(sprintf('Stimulus (N=%d, T=%.2g s, NP=%d periods)',N, N/fs,NP));
xlabel('Time (s)'); ylabel('Level');
plot((1:Ns)/fs,y(:,1),'b');
plot((1:Ns)/fs,y(:,2),'r');
legend('L','R','location','southeast');
str=sprintf('%6.1f dBFS | %6.1f dBFS', ...
  20*log10(max(abs(y(1:N,1)))),20*log10(max(abs(y(1:N,2)))));
text(0,0.9,str,'units','normalized','fontsize',14);
drawnow;
end %if figStimulus

if exist('figInput'),
set(0,'CurrentFigure',figInput); % old way, figure(figInput), causes flicker!
clf; hold on; grid on;
xlim([0 ns/fs]); ylim([-1.05 1.05]);
title(sprintf('Capture (N=%d, T=%.2g s)',N, N/fs));
xlabel('Time (s)'); ylabel('Level');
plot((1:ns)/fs,z(:,1),'b');
plot((1:ns)/fs,z(:,2),'r');
legend('ref','dut','location','southeast');
plot([t1 t1], [-1.1 1.1],'k-','LineWidth',2);
plot([t2 t2], [-1.1 1.1],'k-','LineWidth',2);
str=sprintf('%6.1f dBFS | %6.1f dBFS', ...
  20*log10(max(abs(ref))),20*log10(max(abs(dut))));
text(0,0.9,str,'units','normalized','fontsize',14);
drawnow;
end %if figInput


% Do FFT, plot power spectra and transfer functions
REF=fft(ref);
DUT=fft(dut);
halfN=floor(N/2);
halfN=floor(N/2);
f=((1:halfN)-1)/halfN*(fs/2);
H=DUT./REF;

% Censor transfer function, only output spectrum where there is input signal
dBmin=-90; % minimum level of signal for transfer function
censor_idx= 20*log10(abs(REF)/N) < dBmin;
H(censor_idx)=0;
h=ifft(H); % Impulse response, no censoring

if exist('figSpectrum'),
%figure(figSpectrum); clf; % This flickers in Matlab!  Use code below.
set(0,'CurrentFigure',figSpectrum); clf;
semilogx(f(2:halfN),20*log10((2/N)*abs(REF(2:halfN))),'b'); hold on; grid on;
semilogx(f(2:halfN),20*log10((2/N)*abs(DUT(2:halfN))),'r');
title(sprintf('Spectrum (N=%d)',N));
xlabel('Log Freq (Hz)');
ylabel('20*log10 |X|');
xlim([fs/N fs/2]); ylim([-dB1-dB2 0]); %ylim([-53 -50]); %
legend('REF','DUT','location','southeast');
% dB rms is 20*log10( std(sig) ).  standard deviation
str=sprintf('%6.1f dBrms | %6.1f dBrms', ...
  20*log10(std(ref)),20*log10(std(dut)));
text(0,0.9,str,'units','normalized','fontsize',14);
drawnow;
end %if figSpectrum

if exist('figPhase'),
set(0,'CurrentFigure',figPhase); clf;
semilogx(f(2:halfN),mod(angle(H(2:halfN))+pi,2*pi),'k.'); hold on; grid on;

title(sprintf('Spectrum (Phase) (N=%d)',N));
xlabel('Log Freq (Hz)');
ylabel('angle(X)');
legend('REF','DUT','location','southeast');
xlim([fs/N fs/2]); ylim([-pi pi]);
drawnow;
end %figPhase

if exist('figTF'),
set(0,'CurrentFigure',figTF); clf;
semilogx(f(2:halfN),20*log10(abs(H(2:halfN))),'k.'); grid on;
title('Transfer Function');
% transfer function is dots
xlabel('Log Freq (Hz)');
ylabel('20*log10 |DUT./REF|');
axis tight;
drawnow;
end %if figTF

if exist('figImpulse'),
set(0,'CurrentFigure',figImpulse); clf; hold on; grid on;
xlim([0 N/fs]); ylim([-1.05 1.05]);
title(sprintf('Impulse response'));
xlabel('Time (s)'); ylabel('Level');
plot((1:N)/fs,h);
drawnow;
end %if figImpulse

if exist('figScatter'),
set(0,'CurrentFigure',figScatter); clf;
scatter(ref,dut,1);
xlim([-1.05 1.05]); ylim([-1.05 1.05]);
axis square
title('Scatter plot');
xlabel('REF');
ylabel('DUT');
drawnow;
end % if figScatter

end % if RUN

if get(rWRITE,'value'),
[fn]=uiputfile('*.wav','Output file (stereo WAV)');
audiowrite(fn,[ref,dut],fs);
set(rWRITE,'value',0);
set(rRUN,'value',0);
end %if WRITE

if get(rEXIT,'value') break; end;

if ~get(rRUN,'value') pause(0.1); end

end %while
close all;
